//
//  SWGrid.m
//  White Detector Grid
//
//  Created by David Langford on 6/05/09.
//  Copyright 2009 Sentient Worlds. All rights reserved.
//
#import <OpenGL/CGLMacro.h>
#import "SWGrid.h"
#import "SWCell.h"
#import "Histogram.h"

@implementation SWGrid

@synthesize numberOfColumns
			, numberOfRows
			, widthPixels
			, heightPixels
			, cells
			, widthNormalised
			, heightNormalised
			, xOffsetNormalised
			, yOffsetNormalised	
			, xOffsetPixels
			, yOffsetPixels;

- (id) initWithNumberOfColumns:(NSNumber*)columns 
				  numberOfRows:(NSNumber*)rows
					imageToUse:(id<QCPlugInInputImageSource>)theImage
				   withContext:(id<QCPlugInContext>)theContext
				  widthPercent:(NSNumber*)theWidthPercent
				 heightPercent:(NSNumber*)theHeightPercent
				xOriginPercent:(NSNumber*)theXOriginPercent
				yOriginPercent:(NSNumber*)theYOriginPercent
{
	// Validate arguments
	if ([columns floatValue] < 1 // need columns and rows
		|| [rows floatValue] < 1
		|| !theImage // need an image
		|| [theWidthPercent floatValue] <= 0 // Need some width and height
		|| [theHeightPercent floatValue] <= 0
		|| [theXOriginPercent floatValue] < 0 // Can't start the grid in negative
		|| [theYOriginPercent floatValue] < 0)
	{
		[self release];
		return nil;
	}
	
	// Cap arguments if needed
	float widthPercentTemp = [theWidthPercent floatValue];
	float heightPercentTemp = [theHeightPercent floatValue];
	if (widthPercentTemp + [theXOriginPercent floatValue] > 100)
		widthPercentTemp = 100 - [theXOriginPercent floatValue];
	if (heightPercentTemp + [theYOriginPercent floatValue] > 100)
		heightPercentTemp = 100 - [theYOriginPercent floatValue];
	
	// initialise
	if (self = [super init])
	{
		self.numberOfColumns = [columns copy];
		self.numberOfRows = [rows copy]; 
		self.widthNormalised = [[NSNumber alloc] initWithFloat:widthPercentTemp/100.0];
		self.heightNormalised = [[NSNumber alloc] initWithFloat:heightPercentTemp/100.0];
		self.xOffsetNormalised = [[NSNumber alloc] initWithFloat:[theXOriginPercent floatValue]/100.0];
		self.yOffsetNormalised = [[NSNumber alloc] initWithFloat:[theYOriginPercent floatValue]/100.0];
		self.cells = [self generateCellsUsingImage:theImage andContext:theContext];
	}
	
	return self;
}

// Free any memory we have used
- (void) dealloc
{
	// Release the cells and the NSNumbers
	[self.numberOfColumns release];
	[self.numberOfRows release];
	[self.widthNormalised release];
	[self.heightNormalised release];
	[self.xOffsetNormalised release];
	[self.yOffsetNormalised release];
	[self.cells release];
	[super dealloc];
}

- (NSArray*) generateCellsUsingImage:(id<QCPlugInInputImageSource>)imageToUse
						  andContext:(id<QCPlugInContext>)theContext
{
	// Not to sure what this does exactly, but it gave me a couple more fps
	CGColorSpaceRef					colorSpace;
	colorSpace = (CGColorSpaceGetModel([imageToUse imageColorSpace]) == kCGColorSpaceModelRGB ? [imageToUse imageColorSpace] : [theContext colorSpace]);

	// Get a buffer representation from the image (This is an expensive call, keep outside loop)
	if (![imageToUse lockBufferRepresentationWithPixelFormat:QCPlugInPixelFormatARGB8
												  colorSpace: colorSpace
												   forBounds:[imageToUse imageBounds]])
		
		return NO;
	
	// store the width and height of the image, then find cell width and height. Also find the offset
	self.widthPixels = [[NSNumber alloc] initWithFloat:[imageToUse bufferPixelsWide]*[self.widthNormalised floatValue]];
	self.heightPixels = [[NSNumber alloc] initWithFloat:[imageToUse bufferPixelsHigh]*[self.heightNormalised floatValue]];
	self.xOffsetPixels = [[NSNumber alloc] initWithFloat:[imageToUse bufferPixelsWide]*[self.xOffsetNormalised floatValue]];
	self.yOffsetPixels = [[NSNumber alloc] initWithFloat:[imageToUse bufferPixelsWide]*[self.yOffsetNormalised floatValue]];
	NSNumber* cellWidth = [NSNumber numberWithFloat:[self.widthPixels floatValue]/[self.numberOfColumns intValue]];
	NSNumber* cellHeight = [NSNumber numberWithFloat:[self.heightPixels floatValue]/[self.numberOfRows intValue]];
	
	// Generate weight for each cell and store the values
	NSMutableArray* newCells = [NSMutableArray arrayWithCapacity:[self.numberOfColumns intValue]];
	Boolean error = NO;
	for (NSInteger x=0; x<[self.numberOfColumns intValue]; x++)
	{
		// This will hold an row of cells
		NSMutableArray* rowOfNewCells = [[NSMutableArray alloc] init];
		
		// Loop through each cell in this array
		for (NSInteger y=0; y<[self.numberOfRows intValue]; y++)
		{
			// Create the initial position for this cell
			NSNumber* xPosition = [NSNumber numberWithFloat:x*[cellWidth floatValue]+[self.xOffsetPixels floatValue]];
			NSNumber* yPosition = [NSNumber numberWithFloat:y*[cellHeight floatValue]+[self.yOffsetPixels floatValue]];
			
			// Create the cell
			SWCell* newCell = [[SWCell alloc] initWithXPositionAt:xPosition
													  yPositionAt:yPosition
													widthInPixels:cellWidth
												   heightInPixels:cellHeight];
			
			// Tell the cell to generate its weight
			Boolean success = [newCell generateWeightFromImage:imageToUse 
													 inContext:theContext];
			
			// Record if error happened
			if (!success)
				error = YES;
			
			// Store this data in an array that we can return
			[rowOfNewCells insertObject:[newCell asNSArray]
								atIndex:y];
			
			// We don't need this cell anymore or the numbers
			[newCell release];
		}
		// Store this row of cells
		[newCells insertObject:rowOfNewCells
							  atIndex:x];
	}
	
	// Release the buffer representation and handle errors 
	[imageToUse unlockBufferRepresentation];
	[self.widthPixels release];
	[self.heightPixels release];
	[self.xOffsetPixels release];
	[self.yOffsetPixels release];
	
	// Return nothing if there was an error
	if (error)
		return nil;
	
	// Return all the cells
	return newCells;
}



// Returns an NSArray representing the grid and its cells
- (NSArray*) asNSArray
{
	return self.cells;
}

// Return a description of this grid
- (NSString*) description
{
	NSMutableString* description = [[NSMutableString alloc] initWithString:@"Grid Instance"];
	[description appendFormat:@"\n\tNumber of columns: %i", [self.numberOfColumns intValue]];
	[description appendFormat:@"\n\tNumber of rows: %i", [self.numberOfRows intValue]];
	[description appendFormat:@"\n\tWidth in pixels: %i", [self.widthPixels intValue]];
	[description appendFormat:@"\n\tHeight in pixels: %i", [self.heightPixels intValue]];
	[description appendFormat:@"\n\tNumber of cells: %i by %i", [cells count], [[cells lastObject] count]];
	return description;
}

@end
